#!/usr/bin/bash

SHARED_CONFIG="/tmp/os2borgerpc.conf"

# Current directory
DIR=$(dirname "${BASH_SOURCE[0]}")

while true; do
    fatal() {
        echo "Critical error. Halting registration:" "$@"
        while true; do
            echo "[R]estart or [C]ancel registration?"
            stty -echo
            read -rn 1 VALUE
            stty echo
            case "$VALUE" in
                r|R)
                    rm -f "$SHARED_CONFIG"
                    return 0 ;;
                c|C)
                    return 1 ;;
            esac
        done
    }

    # Get hold of config parameters, connect to admin system.

    if [ "$(id -u)" != "0" ]
    then
        fatal "This program must be run as root" && continue || exit 1
    fi

    echo ""

    # The following config parameters are needed to finalize the
    # installation:
    # - hostname
    #   Prompt user for new host name
    echo "Please enter a new name for this computer." \
         "The name must have a length of 1-63 characters," \
         "and valid characters are a-z, A-Z, 0-9 and hyphen (-):"
    # https://www.man7.org/linux/man-pages/man7/hostname.7.html
    read -r NEW_COMPUTER_NAME
    while [[ ! "$NEW_COMPUTER_NAME" =~ ^[0-9a-zA-Z][0-9a-zA-Z-]{1,63}$ ]]; do
        echo "Invalid computer name. Try again:"
        read -r NEW_COMPUTER_NAME
    done

    # Idea: Allow uppercase in the computername due to popular demand,
    # but lowercase it so it's a valid hostname which is case insensitive
    NEW_HOSTNAME=$(echo "$NEW_COMPUTER_NAME" | tr '[:upper:]' '[:lower:]')

    echo "$NEW_HOSTNAME" > /etc/hostname
    set_os2borgerpc_config hostname "$NEW_HOSTNAME"
    hostname "$NEW_HOSTNAME"
    sed --in-place /127.0.1.1/d /etc/hosts
    sed --in-place "2i 127.0.1.1	$NEW_HOSTNAME" /etc/hosts

    echo ""

    # - site

    echo "Enter your site UID:"
    read -r SITE
    if [[ -z "$SITE" ]]
    then
        fatal "The computer cannot be registered without a site" && continue || exit 1
    fi


    # - mac
    #   Get the mac-address
    set_os2borgerpc_config mac "$(ip addr | grep link/ether | awk 'FNR==1{print $2}')"

    echo ""


    # - admin_url
    unset ADMIN_URL
    if [[ -z "$ADMIN_URL" ]]
    then
        ADMIN_URL="https://os2borgerpc-admin.magenta.dk"
        echo "Press <ENTER> to register with the following admin portal: $ADMIN_URL."
        echo "Alternatively, type in your URL to another instance of the admin portal here:"
        read -r NEW_ADMIN_URL
        if [[ -n "$NEW_ADMIN_URL" ]]
        then
            ADMIN_URL="$NEW_ADMIN_URL"
        fi
    fi
    set_os2borgerpc_config admin_url "$ADMIN_URL"

    # - set additional config values
    OS_NAME=$(lsb_release --id --short)
    set_os2borgerpc_config os_name "$OS_NAME"

    OS_RELEASE=$(lsb_release --release --short)
    set_os2borgerpc_config _os_release "$OS_RELEASE"

    # xargs is there to remove leading and trailing spaces
    PC_MODEL=$(dmidecode --type system | grep Product | cut --delimiter : --fields 2 | xargs)
    [ -z "$PC_MODEL" ] && PC_MODEL="Identification failed"
    PC_MODEL=${PC_MODEL:0:100}
    set_os2borgerpc_config pc_model "$PC_MODEL"

    PC_MANUFACTURER=$(dmidecode --type system | grep Manufacturer | cut --delimiter : --fields 2 | xargs)
    [ -z "$PC_MANUFACTURER" ] && PC_MANUFACTURER="Identification failed"
    PC_MANUFACTURER=${PC_MANUFACTURER:0:100}
    set_os2borgerpc_config pc_manufacturer "$PC_MANUFACTURER"

    CPUS_BASE_INFO="$(dmidecode --type processor | grep Version | cut --delimiter ':' --fields 2 | xargs)"
    CPUS_BASE_INFO=${CPUS_BASE_INFO:0:100}
    CPU_CORES="$(grep ^"core id" /proc/cpuinfo | sort -u | wc -l)"
    CPU_CORES=${CPU_CORES:0:100}
    CPUS="$CPUS_BASE_INFO - $CPU_CORES physical cores"
    [ -z "$CPUS" ] && CPUS="Identification failed"
    set_os2borgerpc_config pc_cpus "$CPUS"

    RAM="$(LANG=c lsmem | grep "Total online" | cut --delimiter ':' --fields 2 | xargs)"
    [ -z "$RAM" ] && RAM="Identification failed"
    RAM=${RAM:0:100}
    set_os2borgerpc_config pc_ram "$RAM"

    # OK, we got the config.
    # Do the deed.
    if ! os2borgerpc_register_in_admin "$NEW_COMPUTER_NAME" "$SITE"; then
        fatal "Registration failed" && continue || exit 1
    fi

    # Now setup cron job
    if [[ -f $(command -v jobmanager) ]]
    then
        # Randomize cron job to avoid everybody hitting the server the same minute
        "$DIR/randomize_jobmanager.sh" 5 > /dev/null
    fi

    break
done
